/*
						+--------------------------------+
                        |                                |
                        |     *** TX2 Compiller ***      |
                        |                                |
                        |        L. Polok  2002         |
                        |                                |
                        |         ImgLoader.cpp          |
                        |                                |
						+--------------------------------+
*/

#define	BMP_MODE_ABS    0
#define BMP_MODE_RLE    1

#define	BI_RGB	0L
#define	BI_RLE8	1L
#define	BI_RLE4	2L

#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "ImgLoader.h"

CImgLoader::CImgLoader()
{

}

CImgLoader::~CImgLoader()
{

}

int CImgLoader::DetectImg_Type(char *filename)
{
	char buff[3];
	FILE *fr;

	if((fr = fopen(filename, "rb")) == NULL)
		return IT_UNKNOWN;
	fread(buff, sizeof(buff), 1, fr);
	fclose(fr);
	// reads first three bytes of file

	if(buff[0] == 'B' && buff[1] == 'M')
		return IT_BITMAP;
	// bitmap

	if(buff[0] == 'G' && buff[1] == 'I' && buff[2] == 'F')
		return IT_GIF;
	// gif

	if(buff[0] == 0x0a && (buff[1] == 0 ||
	   buff[1] == 2 || buff[1] == 3 || buff[1] == 5))
		return IT_PCX;
	// pcx

	return IT_UNKNOWN;
}

BMP *CImgLoader::Load_GIF(char *filename)
{
	return NULL;
}

BMP *CImgLoader::Load_BMP(char * filename)
{
	BITMAPINFOHEADER bmih;
	BITMAPFILEHEADER bmfh;
	FILE *BMPfp;
	BMP *tmp;

	if((BMPfp = fopen(filename, "rb")) == NULL)
		return NULL;
	// opens file

	fread(&bmfh, sizeof(bmfh), 1, BMPfp);
	fread(&bmih, sizeof(bmih), 1, BMPfp);
	// reads header

	if(bmfh.bfType != ('B' | ('M' << 8))) {
		fclose(BMPfp);
		return NULL;
	}
	// checks that file is bitmap

	if(bmih.biBitCount == 8 && bmih.biCompression == BI_RGB) {
		if(Read256ColorBMP_RGB(&bmih, BMPfp) == 0) {
			fclose(BMPfp);
			return NULL;
		}
	} else if(bmih.biBitCount == 8 && bmih.biCompression == BI_RLE8) {
		if(Read256ColorBMP_RLE8(&bmih, BMPfp) == 0) {
			fclose(BMPfp);
			return NULL;
		}
	} else
		return NULL;
	// reading bitmap

	fclose(BMPfp);

	if((tmp = (BMP*) malloc(sizeof(BMP))) == NULL)
		return NULL;
	tmp->Width = bmih.biWidth;
	tmp->Height = bmih.biHeight;
	tmp->bytes = bytes;
	// saving data to common format

	return tmp;
}

BMP *CImgLoader::Load_PCX(char *filename)
{
	unsigned char *decodedBuffer;
    unsigned char *imageBuffer;
	unsigned short numOfColors;
	PCXVGAPALETTE PcxVgaPal;
    unsigned short xSize;
    unsigned short ySize;
    short scanLineLength;
    PCXHEADER pcxHead;
    FILE *fpPcxIn;
    char int8[1];
    short ret;
	BMP *tmp;
	int i, n;

    // Open the input PCX image file
    if((fpPcxIn = fopen(filename, "rb")) == (FILE *) NULL)
		return NULL;
    // Read the PCX image file header information
    ReadPcxHeader(&pcxHead, fpPcxIn);
    // Check the identification byte value
    if(pcxHead.Id != 0x0A)
        return NULL;
    // Calculate size of image in pixels and scan lines
    xSize = pcxHead.Xmax - pcxHead.Xmin + 1;
    ySize = pcxHead.Ymax - pcxHead.Ymin + 1;
    // Do some more calculations
    numOfColors = (1 << (pcxHead.BitsPixelPlane * pcxHead.NumberOfPlanes));
    scanLineLength = pcxHead.NumberOfPlanes * pcxHead.BytesLinePlane;
    // Create an uncompressed scanline buffer
    if((decodedBuffer = (unsigned char *) malloc(scanLineLength)) == NULL)
        return NULL;
	if((imageBuffer = (unsigned char *) malloc(scanLineLength * ySize)) == NULL)
        return NULL;

    for(i = 0; i < ySize; i ++) {
        // Read scan line and decode it
        if((ret = PcxDecodeScanLine(decodedBuffer, scanLineLength, fpPcxIn))
          != scanLineLength) {
            printf("PCXCODE: Bad scan line length (was %d should be %d)\n",
              ret, scanLineLength);
        }
		memcpy(&imageBuffer[i * xSize], decodedBuffer, scanLineLength);
		// copy data to image buffer
    }

    if(pcxHead.Version == 5) {
        fseek(fpPcxIn, -769L, SEEK_END);
		fread(int8, 1, 1, fpPcxIn);
        if(int8[0] == 0x0C) {
            // Read the VGA palette
            fread(&PcxVgaPal.VgaPalette, 1, 768, fpPcxIn);
			fclose(fpPcxIn);

			if((tmp = (BMP*) malloc(sizeof(BMP))) == NULL)
				return NULL;
			if((tmp->bytes = (unsigned __int32*) malloc(4 * xSize * ySize)) == NULL)
				return NULL;
			// lock memotry for bitmap

			tmp->Width = xSize;
			tmp->Height = ySize;
			for(i = 0; i < xSize * ySize; i ++) {
				n = imageBuffer[i];
				tmp->bytes[i] = PcxVgaPal.VgaPalette[n * 3] +
					(PcxVgaPal.VgaPalette[n * 3 + 1] << 8) +
					(PcxVgaPal.VgaPalette[n * 3 + 2] << 16);
			}
			// creating ttruecolor bitmap

			free((void*)imageBuffer);
			free((void*)decodedBuffer);
			// freeing up memory
			
			return tmp;
        }
    }
    fclose(fpPcxIn);

	return NULL;
}

BMP	*CImgLoader::Load_TrueColor_BMP(char *filename)
{
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	BMP *tmp;
	unsigned char *lpbBuf;
	unsigned int Width, Height;
	unsigned int BuffLen;
	FILE *BMPfp;
	int	i, j;
	
	if((BMPfp = fopen(filename, "rb")) == NULL)
		return NULL;
	// Opens file

	fread(&bmfh, sizeof(bmfh), 1, BMPfp);
	fread(&bmih, sizeof(bmih), 1, BMPfp);
	// Reads infoheaders

	if(bmfh.bfType != ('B'|('M'<<8)))
		return NULL; 
	// format 'BM'

	if(bmih.biBitCount == 24 && bmih.biCompression == BI_RGB) {
		if((tmp = (BMP*) malloc(sizeof(BMP))) == NULL)
			return NULL;
		// Allocation of memory

		Width = (unsigned)bmih.biWidth;
		Height = (unsigned)bmih.biHeight;
		// Width & Height

		tmp->Width = Width;
		tmp->Height = Height;
		if((tmp->bytes = (unsigned __int32*) malloc(
			Width*Height*sizeof(unsigned __int32))) == NULL) return NULL;
		// Bitmap was prepared to load

		BuffLen = ((Width * 3 + 3) >> 2) << 2;
		// Must be aligned to 32-bit chunks

		if((lpbBuf = (unsigned char*)malloc(BuffLen)) == NULL)
			return NULL;
		// Allocation of helper memory

		for(i = Height - 1; i >= 0; i--) {
			if((j = fread(lpbBuf, 1, BuffLen,
			   BMPfp)) != (signed)BuffLen) {
				return NULL;
			} // Reads byte

			for(j = 0; j < (signed)Width; j ++) {
				tmp->bytes[j + Width * i] = (lpbBuf[j * 3] << 16) + 
					(lpbBuf[j * 3 + 1] << 8) + lpbBuf[j* 3 + 2];
			}
			// write to buffer
		}
	} else
		return NULL;

	free(lpbBuf);
	return tmp;
}

int	CImgLoader::Read256ColorBMP_RGB(BITMAPINFOHEADER *bmih, FILE *BMPfp)
{
    unsigned int color_cnt = 256;
	unsigned int buff_len;
	RGBQUAD palette[256];
	unsigned int Height;
    unsigned int Width;
	unsigned char *lpbBuf;
    int i, j;

	Width = (unsigned)bmih->biWidth;
	Height = (unsigned)bmih->biHeight;
	buff_len = ((Width + 3) >> 2) << 2;
	// compute sizes of buffers

    if((lpbBuf = (unsigned char *)malloc(buff_len)) == NULL)
		return 0;
	if((bytes = (unsigned __int32*)malloc(Width * Height * 4)) == NULL)
		return 0;
	// lock memory for buffers

    fread(palette, sizeof(RGBQUAD), color_cnt, BMPfp);
	// read pallete

    for(i = 0; i < (signed)color_cnt; i ++) {
        paleta[i] = palette[i].rgbBlue + (palette[i].rgbGreen << 8) +
	 	    (palette[i].rgbRed << 16);
    }
	// convert palette to truecolor

    for(i = Height - 1; i >= 0; i --) {
        if(fread(lpbBuf, 1, buff_len, BMPfp) != buff_len) {
	        free(lpbBuf);
	        return 0;
        }
        for(j = 0; j < (signed)Width; j ++)
	        bytes[j + Width * i] = paleta[lpbBuf[j]];
	}
	// reads bitmap

	free((void*)lpbBuf);
	// freeing buffer

    return 1;
}

int	CImgLoader::Read256ColorBMP_RLE8(BITMAPINFOHEADER *bmih, FILE *BMPfp)
{
	unsigned int color_cnt = 256;
	unsigned char *lpbBuf, bBuf;
    unsigned char DummyCount;
	unsigned int buff_len;
	RGBQUAD palette[256];
	unsigned int Height;
	unsigned char Count;
    unsigned int Width;
    int ReadCode, Mode;
	int i, j;

	Width = (unsigned)bmih->biWidth;
	Height = (unsigned)bmih->biHeight;
	buff_len = ((Width + 3) >> 2) << 2;
	// compute size of buffers

	if((bytes = (unsigned __int32*)malloc(Width * Height * 4)) == NULL)
		return 0;
    if((lpbBuf = (unsigned char*)malloc(buff_len)) == NULL)
		return 0;
	// lock memory for buffers

    fread(palette, sizeof(RGBQUAD), color_cnt, BMPfp);
	// reads pallete

    for(i = 0; i < (signed)color_cnt; i ++) {
        paleta[i] = palette[i].rgbBlue + (palette[i].rgbGreen << 8) +
			(palette[i].rgbRed << 16);
	}
	// converts palette to truecolor

    ReadCode = 1;

    for(i = Height - 1; i >= 0; ) {
        if(ReadCode == 1) {
	        bBuf = fgetc(BMPfp);
	        ReadCode = 0;
	        if(bBuf == 0) {
	            bBuf = fgetc(BMPfp);
				switch(bBuf) {
				case 0x00:
					j = Width;
					ReadCode = 1;
					break;
				case 0x01:
					j = Width;
					i = 0;
					break;
				case 0x02:
					ReadCode = 1;
					bBuf = fgetc(BMPfp);
					j += bBuf;
					bBuf = fgetc(BMPfp);
					i -= bBuf;
					if(j >= (signed)Width) {
						j -= Width;
						i --;
					}
					break;
				default:
					Mode = BMP_MODE_ABS;
					Count = bBuf;
					break;
				}
			} else {
				Mode = BMP_MODE_RLE;
				Count = bBuf;
			}
		} else {
			if(Mode == BMP_MODE_ABS) {
				DummyCount = Count % 2;
				while(Count -- > 0) {
					bBuf = fgetc(BMPfp);
					if(j == (signed)Width) {
						j = 0;
						i --;
					}
					bytes[j + Width * i] = paleta[bBuf];
					j ++;
				}
				if(DummyCount != 0)
					bBuf = fgetc(BMPfp);
			} else {
				bBuf = fgetc(BMPfp);
				while(Count -- > 0) {
					if(j == (signed)Width) {
						j = 0;
						i --;
					}
					bytes[j + Width * i] = paleta[bBuf];
					j ++;
				}
			}
			ReadCode = 1;
		}
	}
	// reading bitmap

	return 1;
}

void CImgLoader::ReadPcxHeader(PCXHEADER *PcxHead, FILE *FpPcx)
{
	short int16[1];
	char int8[1];

	// reading header
	fread(int8, sizeof(int8), 1, FpPcx);
    PcxHead->Id = int8[0];
	fread(int8, sizeof(int8), 1, FpPcx);
    PcxHead->Version = int8[0];
	fread(int8, sizeof(int8), 1, FpPcx);
    PcxHead->Format = int8[0];
	fread(int8, sizeof(int8), 1, FpPcx);
    PcxHead->BitsPixelPlane = int8[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->Xmin = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->Ymin = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->Xmax = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->Ymax = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->Hdpi = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->Vdpi = int16[0];

    /* Read the EGA Palette */
    fread(PcxHead->EgaPalette, sizeof(char), 48, FpPcx);

	fread(int8, sizeof(int8), 1, FpPcx);
    PcxHead->Reserved = int8[0];
	fread(int8, sizeof(int8), 1, FpPcx);     
    PcxHead->NumberOfPlanes = int8[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->BytesLinePlane = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx);
    PcxHead->PaletteInfo = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx);  
    PcxHead->HScreenSize = int16[0];
	fread(int16, sizeof(int16), 1, FpPcx); 
    PcxHead->VScreenSize = int16[0];

    /* Read the reserved area at the end of the header */
	fread(PcxHead->Filler, sizeof(char), 54, FpPcx);
}

short CImgLoader::PcxDecodeScanLine(unsigned
	char *DecodedBuffer, int BufferSize, FILE *FpPcx)
{
	static char rcount = 0;
    static char rvalue = 0;
	short total = 0;
    int  index = 0;
    char  byte;

    do {
        for(total += rcount; rcount && index < BufferSize; rcount -- , index ++)
            DecodedBuffer[index] = rvalue;
		// decoding scanline

        if(rcount) {
            total -= rcount;
            return(total);
        } // return count of decoded pixels

		fread(&byte, 1, 1, FpPcx);
        if((byte & 0xC0) == 0xC0) {
            rcount = byte & 0x3F;
            fread(&rvalue, 1, 1, FpPcx);
        } else {
            rcount = 1;
            rvalue = byte;
        }
		// get and decode next pixel
    } while(index < BufferSize);

	// return count of decoded pixels
    return(total);
}

int	CImgLoader::Save_TrueColor_BMP(char *filename, BMP *bmp)
{
	unsigned long ByteCount;
	BITMAPFILEHEADER bmfh;
	BITMAPINFOHEADER bmih;
	unsigned char *lpbBuf;
	unsigned __int32 bBuf;
	unsigned int BuffLen;
	int i, j, x;
	FILE *BMPfp;

	BuffLen = ((3 * bmp->Width + 3) >> 2) << 2;
	lpbBuf =  (unsigned char*) malloc(BuffLen);
	// compute lenght of line-buffer

	if((BMPfp = fopen(filename, "wb")) == NULL)
		return 0;
	// opens file

	bmfh.bfType = ('B'|('M'<<8)); // "BM"
	bmfh.bfSize = 0;              // necessary to modify
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) +
		sizeof(BITMAPINFOHEADER);
	bmih.biSize = sizeof(BITMAPINFOHEADER);
	bmih.biWidth = bmp->Width;
	bmih.biHeight = bmp->Height;
	bmih.biPlanes = 1;
	bmih.biBitCount = 24;	      // True Color
	bmih.biCompression = BI_RGB;
	bmih.biSizeImage = 0;         // necessary to modify
	bmih.biXPelsPerMeter = 96;
	bmih.biYPelsPerMeter = 96;
	bmih.biClrUsed = 0;
	bmih.biClrImportant = 0;
	// create header

	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, BMPfp);
	fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, BMPfp);
	// writing header

	ByteCount = 0;
	for(i = bmp->Height - 1; i >= 0; i --) {
		for(j = 0, x = 0; j < bmp->Width; j ++, x ++) {
			bBuf = bmp->bytes[i * bmp->Width + j];
			lpbBuf[x * 3] = (char)((bBuf >> 16) & 0xff);
			lpbBuf[x * 3 + 1] = (char)((bBuf >> 8) & 0xff);
			lpbBuf[x * 3 + 2] = (char)(bBuf & 0xff);
			// saved in BGR, not RGB
			ByteCount += 3;
		}

		if(fwrite(lpbBuf, 1, BuffLen, BMPfp) != BuffLen) {
			free(lpbBuf);
			return 0;
		}
	}

	// repair header
	bmih.biSizeImage = ByteCount;
	bmfh.bfSize = sizeof(BITMAPFILEHEADER) +
	   ByteCount + sizeof(BITMAPINFOHEADER);
	fseek(BMPfp, 0, SEEK_SET);
	fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, BMPfp);
	fwrite(&bmih, sizeof(BITMAPINFOHEADER), 1, BMPfp);

	fclose(BMPfp);

	free(lpbBuf);

	return 1;
}