/*
						+--------------------------------+
                        |                                |
                        |     *** TX2 Compiller ***      |
                        |                                |
                        |        L. Polok  2002         |
                        |                                |
                        |        TxCompiller.cpp         |
                        |                                |
						+--------------------------------+
*/


#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "ImgLoader.h"
#include "FileList.h"
#include "IntTable.h"
#include "TxCompiller.h"
#include "Function.h"

typedef struct {
	char type[3];
	short version;
	short pal_num;
	short mip_num;
	short frm_num;
	short compression;
	short TWidth, THeight;
	short tex_type;
} TEXTURE_HEADER;

CTxCompiller::CTxCompiller()
{
	//
	ildr = new CImgLoader;
	imgs = new CFileList;
	pal = new CIntTable;
	// create potwors
}

CTxCompiller::~CTxCompiller()
{
	//
	delete pal;
	delete ildr;
	delete imgs;
	// freeing memory
}

void CTxCompiller::Mem_Err()
{
	printf(" ERROR : Not enough physical memory !\n");
	getchar();
	// print not-enough-mem message and waits for enter
}

void CTxCompiller::Std_Err(char *err)
{
	printf(" ERROR : %s\n", err);
	getchar();
	// print an error message and waits for enter
}

int CTxCompiller::CreateTexture(char *srcname, int compress)
{
	__int16 int16[1];
	__int8 int8[1];
	BMP **bmp_list;
	BMP **mip_list;
	char dest[256];
	int pix_cnt;
	int bmp_cnt;
	int mip_cnt;
	int i, j, c;
	int r_size;
	int e_size;
	int *func;
	FILE *fw;
	int k;

	CFunction *f;

	/*
	int fn[] = {0, 0, 0, 1,
	            0, 1, 1, 0,
				1, 1, 1, 0,
				0, 0, 0, 1};

	f = new CFunction(16, fn);
	f->Print();
	delete f;
	*/ // testing class CFunction

	strcpy(dest, srcname);
	if(strrchr(dest, '.') != NULL)
		strcpy(strrchr(dest, '.'), ".tx2");
	else
		strcat(dest, ".tx2");
	// create texture destination filename

	/* ----------------------------------------------------------*/

	switch(ildr->DetectImg_Type(srcname)) {
	case IT_BITMAP:
	case IT_PCX:
	case IT_GIF:
		// single file ...
		printf("Single image ...\n");
		if(imgs->AddFile(srcname, 0) == -1) {
			Mem_Err();
			// not enough memory
			return 0;
		}
		// create filelist "by hand"
		break;
	case IT_UNKNOWN:
		printf("Loading filelist ...\n");
		if(!imgs->Load(srcname)) {
			Std_Err("Unknown format !");
			// unable to load filelist
			return 0;
		}
		break;
	default:
		Std_Err("Unknown format !");
		// unknown format
		return 0;
	}
	// create filelist

	/* ----------------------------------------------------------*/

	printf("Locking memory for images (%d)..\n", imgs->FileNum());
	bmp_cnt = imgs->FileNum();
	if((bmp_list = (BMP**) malloc
	   (bmp_cnt * sizeof(BMP*))) == NULL) {
		Mem_Err();
		return 0;
	}
	// bitmap array

	printf("Loading picture #%d ..", 0);
	for(i = 0; i < bmp_cnt; i ++) {
		printf("\rLoading picture #%d ..", i);
		switch(ildr->DetectImg_Type(imgs->File(i)->name)) {
		case IT_BITMAP:
			bmp_list[i] = ildr->Load_TrueColor_BMP(imgs->File(i)->name);
			// loading image
			if(bmp_list[i] != NULL)
				break;
			// true color
			bmp_list[i] = ildr->Load_BMP(imgs->File(i)->name);
			if(bmp_list[i] == NULL) {
				Std_Err("Error loading picture !");
				return 0;
			}
			break;
		case IT_PCX:
			bmp_list[i] = ildr->Load_PCX(imgs->File(i)->name);
			// loading image
			if(bmp_list[i] == NULL) {
				Std_Err("Error loading picture !");
				return 0;
			}
			break;
		case IT_GIF:
			bmp_list[i] = ildr->Load_GIF(imgs->File(i)->name);
			// loading image
			if(bmp_list[i] == NULL) {
				Std_Err("Error loading picture !");
				return 0;
			}
			break;
		default:
			Std_Err("Unknown image format !");
			return 0;
		}
		// loading images
		if(i > 0) {
			if(bmp_list[i]->Width != bmp_list[i - 1]->Width ||
			   bmp_list[i]->Height != bmp_list[i - 1]->Height) {
				Std_Err("Pictures must be same size !");
				// testing picture size
				return 0;
			}
		}
	}
	printf("\rLoading picture #%d ..\n", imgs->FileNum() - 1);
	// loading bitmaps

	//ildr->Save_TrueColor_BMP("potw1.bmp", bmp_list[0]);
	/* ----------------------------------------------------------*/

	pix_cnt = bmp_list[0]->Width * bmp_list[0]->Height;
	printf("Creating palette (%3d cols found, %3d %% done)", 0, 0);
	for(i = 0; i < bmp_cnt; i ++) {
		for(j = 0; j < pix_cnt; j ++) {
			c = pal->Add(bmp_list[i]->bytes[j]);
			if(c == -1) {
				printf("\n");
				Std_Err("Adding color to table failed !");
				return 0;
			}
			if((j & 0x1f) == 0x10) {
				printf("\rCreating palette (%3d cols found"
					", %3d %% done)", pal->Val_Num(),
					((i + (j / 2)) * 100) / (bmp_cnt * pix_cnt));
			}
		} // adding pixels to palette
		for(j = 0; j < pix_cnt; j ++) {
			c = pal->Find(bmp_list[i]->bytes[j]);
			if(c == -1) {
				printf("\n");
				Std_Err("Adding color to texture failed !");
				return 0;
			}
			bmp_list[i]->bytes[j] = (unsigned) c;
			if((j & 0x1f) == 0x10) {
				printf("\rCreating palette (%3d cols found"
					", %3d %% done)", pal->Val_Num(),
					((i + (j + pix_cnt) / 2) * 100) / (bmp_cnt * pix_cnt));
			}
		} // adding pixels to palette
	}
	printf("\rCreating palette (%3d cols found"
		", %3d %% done)\n", pal->Val_Num(), 100);
	// creating palette

	//ildr->Save_TrueColor_BMP("potw.bmp", bmp_list[0]);
	/* ----------------------------------------------------------*/

	i = bmp_list[0]->Width;
	j = bmp_list[0]->Height;

	c = 0;
	while(i) {
		c += i & 1;
		i >>= 1;
	}
	while(j) {
		c += j & 1;
		j >>= 1;
	}

	if(c != 2) {
		Std_Err("Bitmaps must be power-of-two size !");
		Std_Err("Skipping mipmaps !");
		// not power of two bitmaps
		mip_cnt = 0;
	} else {
		i = bmp_list[0]->Width;
		j = bmp_list[0]->Height;
		//
		i = (i < j) ? i : j;
		mip_cnt = (int)(1e-10 + log((float)i) / log((float)2));
		printf(" %d MipMaps to do ..", mip_cnt);
	}

	if((mip_list = (BMP**) malloc(mip_cnt *
	   bmp_cnt * sizeof(BMP*))) == NULL) {
		Mem_Err();
		// not enough memory for mipmaps
		return 0;
	}

	/* ----------------------------------------------------------*/

	for(i = 0; i < bmp_cnt; i ++) {
		printf("\rCreating MipMaps [%2d / %2d : %3d %%]",
			0, mip_cnt, ((0 + i) * 100) / (mip_cnt * bmp_cnt));
		//
		if((mip_list[0 + i * mip_cnt] =
		   this->Create_MipMap(bmp_list[i])) == NULL) {
			Mem_Err();
			return 0;
		}
		// first mipmap
		for(j = 1; j < mip_cnt; j ++) {
			if((mip_list[j + i * mip_cnt] =
			   this->Create_MipMap(mip_list
			   [j - 1 + i * mip_cnt])) == NULL) {
				Mem_Err();
				return 0;
			}
			// creating mipmap
			printf("\rCreating MipMaps [%2d / %2d "
				": %3d %%]", j, mip_cnt,
				((j + i) * 100) / (mip_cnt * bmp_cnt));
		}
	}
	printf("\rCreating MipMaps [%2d / %2d "
		": %3d %%]\n", mip_cnt, mip_cnt, 100);
	// creating mipmaps

	/*
	typedef struct {
		char type[3];
		short version;
		short pal_num;
		short mip_num;
		short frm_num;
		short compression;
		short TWidth, THeight;
		short tex_type;
	} TEXTURE_HEADER;
	*/

	/* ----------------------------------------------------------*/

	TEXTURE_HEADER txhdr;
	unsigned __int32 col;

	pix_cnt = 1 << (int)ceil(log(pix_cnt) / log(2));
	if((func = (int*) malloc(pix_cnt * sizeof(int))) == NULL)
		return 0;
	memset(func, 2, pix_cnt * sizeof(int)); // log. X

	e_size = bmp_list[0]->Width * bmp_list[0]->Height * 4;
	// estimated size of frame in bytes
 
	if((fw = fopen(dest, "wb")) != NULL) {
		printf("Writing texture to HDD ..\n");
		//
		printf("   - Writing header ..\n");
		memcpy(txhdr.type, "tx2", 3);
		txhdr.version = 100;
		txhdr.pal_num = pal->Val_Num();
		txhdr.mip_num = mip_cnt;
		txhdr.frm_num = bmp_cnt;
		txhdr.TWidth = bmp_list[0]->Width;
		txhdr.THeight = bmp_list[0]->Height;
		txhdr.compression = compress;
		txhdr.tex_type = 0; // transparency, masking, etc..
		fwrite(&txhdr, sizeof(txhdr), 1, fw);
		// writing header
		
		printf("   - Writing palette ..\n");
		for(i = 0; i < pal->Val_Num(); i ++) {
			col = pal->Val(i);
			fwrite(&col, sizeof(unsigned __int32), 1, fw);
		}
		// writing palette

		if(compress) {
			printf("   - Compressing image ..\n");

			for(i = 0; i < bmp_cnt; i ++) {
				r_size = 0;
				// real size
				for(k = 0; k < pal->Val_Num(); k ++) {
					for(j = 0; j < bmp_list[0]->Width * bmp_list[0]->Height; j ++) {
						if(bmp_list[i]->bytes[j] == (unsigned)k) {
							func[j] = 1; // log. H
							bmp_list[i]->bytes[j] = -1;
						} else if(bmp_list[i]->bytes[j] == -1)
							func[j] = 2; // log. X
						else
							func[j] = 0; // log. L
						// create function
					}
					f = new CFunction(pix_cnt, func);
					// compress function

					for(j = 0; j < f->Chunk_Len(); j ++) {
						fwrite(&(f->Chunk())[j], 1, 1, fw);
					}
					// write chunk

					r_size += f->Chunk_Len();
					// add chunk len to real size

					printf("\r        - Image %3d, chunk %4d /"
						" %4d, compress = %3d %%", i + 1, k + 1,
						pal->Val_Num(), (int)(((float)(r_size *
						pal->Val_Num() * 100) / (float)k) /
						(float)e_size));

					delete f;
					// free compression class
				}
				printf("\n");
			}
			free((void*)func);
			// writing bitaps

#ifdef WRITE_MIPMAPS
			for(i = 0; i < mip_cnt; i ++) {
				pix_cnt = mip_list[i]->Width * mip_list[i]->Height;
				pix_cnt = 1 << (int)ceil(log(pix_cnt) / log(2));
				if((func = (int*) malloc(pix_cnt * sizeof(int))) == NULL)
					return 0;
				memset(func, 2, pix_cnt * sizeof(int)); // log. X
				for(k = 0; k < pal->Val_Num(); k ++) {
					for(j = 0; j < mip_list[i]->Width * mip_list[i]->Height; j ++) {
						if(mip_list[i]->bytes[j] == (unsigned)k) {
							func[j] = 1; // log. H
							mip_list[i]->bytes[j] = -1;
						} else if(mip_list[i]->bytes[j] == -1)
							func[j] = 2; // log. X
						else
							func[j] = 0; // log. L
						// create function
					}
					printf("\r        - MipMap %3d, chunk %4d / %4d",
						i, k, pal->Val_Num());
					f = new CFunction(pix_cnt, func);
					fwrite(f->Chunk(), 1, f->Chunk_Len(), fw);
					delete f;
				}
				free((void*)func);
				printf("\r        - MipMap %3d, chunk %4d / %4d\n",
					i, pal->Val_Num(), pal->Val_Num());
			}
#endif
			// writing mipmaps
		} else {
			printf("   - Writing image ..\n");

			pix_cnt = bmp_list[0]->Width * bmp_list[0]->Height;
			if(pal->Val_Num() > 256) {
				for(i = 0; i < bmp_cnt; i ++) {
					for(j = 0; j < pix_cnt; j ++) {
						int16[0] = (__int16)bmp_list[i]->bytes[j];
						fwrite(int16, 2, 1, fw);
					}
					printf("\r        - Bitmap %3d\n", i);
				}
			} else {
				for(i = 0; i < bmp_cnt; i ++) {
					for(j = 0; j < pix_cnt; j ++) {
						int8[0] = (__int8)bmp_list[i]->bytes[j];
						fwrite(int8, 1, 1, fw);
					}
					printf("\r        - Bitmap %3d\n", i);
				}
			}
			// writing bitaps

#ifdef WRITE_MIPMAPS
			if(pal->Val_Num() > 256) {
				for(i = 0; i < mip_cnt; i ++) {
					pix_cnt = mip_list[i]->Width * mip_list[i]->Height;
					for(j = 0; j < pix_cnt; j ++) {
						int16[0] = (__int16)mip_list[i]->bytes[j];
						fwrite(int16, 2, 1, fw);
					}
					printf("\r        - MipMap %3d\n", i);
				}
			} else {
				for(i = 0; i < mip_cnt; i ++) {
					pix_cnt = mip_list[i]->Width * mip_list[i]->Height;
					for(j = 0; j < pix_cnt; j ++) {
						int8[0] = (__int8)mip_list[i]->bytes[j];
						fwrite(int8, 1, 1, fw);
					}
					printf("\r        - MipMap %3d\n", i);
				}
			}
			// writing mipmaps
#endif
		}
		fclose(fw);
	} else {
		Std_Err("Cannot open destination file !");
	}

	/* ----------------------------------------------------------*/

	//
	for(i = 0; i < bmp_cnt; i ++)
		free((void*)bmp_list[i]);
	free((void*)bmp_list);
	// freeing memory

	return 1;
}

BMP *CTxCompiller::Create_MipMap(BMP *bmp)
{
	unsigned __int32 col;
	int NWidth, NHeight;
	BMP *tmp_mip;
	int r, g, b;
	int sx, sy;

	NWidth = bmp->Width >> 1;
	NHeight = bmp->Height >> 1;
	// compute width and height

	if((tmp_mip = (BMP*) malloc(sizeof(BMP))) == NULL)
		return NULL;
	if((tmp_mip->bytes = (unsigned __int32*) malloc
	   ((NWidth * NHeight) << 2)) == NULL)
	    return NULL;
	// locking memory

	tmp_mip->Width = NWidth;
	tmp_mip->Height = NHeight;
	// copy params

	for(sx = 0; sx < NWidth; sx ++) {
		for(sy = 0; sy < NHeight; sy ++) {
			col = this->pal->Val(bmp->bytes
				[(sx << 1) + (sy << 1) * bmp->Width]);
			r = (col >> 16) & 0xff;
			g = (col >>  8) & 0xff;
			b = (col      ) & 0xff;
			//
			col = this->pal->Val(bmp->bytes
				[(sx << 1) + 1 + (sy << 1) * bmp->Width]);
			r += (col >> 16) & 0xff;
			g += (col >>  8) & 0xff;
			b += (col      ) & 0xff;
			//
			col = this->pal->Val(bmp->bytes
				[(sx << 1) + ((sy << 1) + 1) * bmp->Width]);
			r += (col >> 16) & 0xff;
			g += (col >>  8) & 0xff;
			b += (col      ) & 0xff;
			//
			col = this->pal->Val(bmp->bytes
				[(sx << 1) + 1 + ((sy << 1) + 1) * bmp->Width]);
			r += (col >> 16) & 0xff;
			g += (col >>  8) & 0xff;
			b += (col      ) & 0xff;
			// compute result color
			r >>= 2;
			g >>= 2;
			b >>= 2;
			//
			tmp_mip->bytes[sx + NWidth * sy] =
				this->Find_NearestColor(r, g, b);
			// find mipped color
		}
	}
	// computing mipmap

	return tmp_mip;
}

int CTxCompiller::Find_NearestColor(int r, int g, int b)
{
	float delta, min_delta;
	int i, n, min_index;
	int ar, ag, ab;

	min_index = 0;
	min_delta = 0x10000;
	// first potwor
	
	n = pal->Val_Num();
	for(i = 0; i < n; i ++) {
		ar = (pal->Val(i) >> 16) & 0xff;
		ag = (pal->Val(i) >>  8) & 0xff;
		ab = (pal->Val(i)      ) & 0xff;
		// actual color
		ar -= r;
		ag -= g;
		ab -= b;
		delta = (float)sqrt((float)(ar * ar + ag * ag + ab * ab));
		// compute delta
		if(delta < min_delta) {
			min_delta = delta;
			min_index = i;
		}
		// if color is closer to rgb, change it
	}
	// searching for close color

	return min_index;
}